package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.experiment.setup.config.Slice;
import be.iminds.ilabt.jfed.ui.cli2.instruction.*;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.library.JFedVersionInfo;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.apache.commons.cli.*;
import org.jetbrains.annotations.Contract;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ExperimenterCommandLineParser
 */
public class ExperimenterCommandLineParser {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimenterCommandLineParser.class);

    private GenericCliArguments genericCliArguments;
    private Instruction instruction;

    private final String[] args;
    private final PrintStream out;
    private final PrintStream err;
    private final InputStream in;
    private final BaseCli baseCli;

    public static final String CHOOSE_PROJECT_AUTOMATICALLY_OPTION = "CHOOSE_AUTOMATICALLY";
    public static final List<String> actionOptions = Arrays.asList("ReloadOS", "Restart", "ConsoleUrl");


    @Nullable
    private <R> R getInYml(@Nonnull Class<R> resultClass, @Nonnull Map instructionMap,
                           String... keys) throws ArgumentException {
        return getInYml(null, resultClass, instructionMap, keys);
    }
    @Nonnull
    private <R> R getInYml(@Nonnull R defaultVal, @Nonnull Map instructionMap,
                           String... keys) throws ArgumentException {
        return getInYml(defaultVal, (Class<R>)defaultVal.getClass(), instructionMap, keys);
    }
    @Nullable
    @Contract("!null,_,_,_ -> !null")
    private <R> R getInYml(@Nullable R defaultVal,
                           @Nonnull Class<R> resultClass,
                           @Nonnull Map instructionMap,
                           String... keys) throws ArgumentException {
        Object cur = instructionMap;
        List<String> seenKeys = new ArrayList<>();
        for (String key : keys) {
            if (!(cur instanceof Map))
                throw new ArgumentException("Unexpected branch type ("+cur.getClass().getName()+") at "+seenKeys+"  (expected Map)");

            cur = ((Map) cur).get(key);

            //if any value along the way is null, return null
            if (cur == null)
                return defaultVal;

            seenKeys.add(key);
        }
        if (!resultClass.isInstance(cur))
            throw new ArgumentException("Unexpected leaf type ("+cur.getClass().getName()+") at "+Arrays.asList(keys)+"  (expected "+resultClass.getName()+")");
        return (R) cur;
    }
    @Nonnull
    private <R> List<R> getListInYml(@Nonnull Class<R> resultClass, @Nonnull Map instructionMap,
                                     String... keys) throws ArgumentException {
        return getListInYml(Collections.emptyList(), resultClass, instructionMap, keys);
    }
    @Nonnull
    private <R> List<R> getListInYml(@Nonnull List<R> defaultVal, @Nonnull Map instructionMap,
                                     String... keys) throws ArgumentException {
        return getListInYml(defaultVal, (Class<R>)defaultVal.getClass(), instructionMap, keys);
    }
    @Nonnull
    private <R> List<R> getListInYml(@Nonnull List<R> defaultVal, @Nonnull Class<R> resultClass, @Nonnull Map instructionMap,
                                     String... keys) throws ArgumentException {
        Object cur = instructionMap;
        List<String> seenKeys = new ArrayList<>();
        for (String key : keys) {
            if (!(cur instanceof Map))
                throw new ArgumentException("Unexpected branch type ("+cur.getClass().getName()+") at "+seenKeys+"  (expected Map)");

            cur = ((Map) cur).get(key);

            //if any value along the way is null, return the defaultVal
            if (cur == null)
                return defaultVal;

            seenKeys.add(key);
        }
        if (!resultClass.isInstance(cur))
            throw new ArgumentException("Unexpected leaf type ("+cur.getClass().getName()+") at "+Arrays.asList(keys)+"  (expected "+resultClass.getName()+")");
        return (List<R>) cur;
    }
    @Nullable
    private <E extends Enum> E getEnumInYml(@Nonnull Map instructionMap,
                                            @Nonnull Class<E> enumClass,
                                            String... keys) throws ArgumentException {
        String val = getInYml(String.class, instructionMap, keys);
        if (val == null)
            return null;
        try {
            Method valueOfMethod = enumClass.getMethod("valueOf", String.class);
            return (E) valueOfMethod.invoke(null, val);
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException("Should never happen for an enum!", e);
        }
    }
    @Nonnull
    private <E extends Enum> E getEnumInYml(@Nonnull Map instructionMap,
                                            @Nonnull E defaultVal,
                                            String... keys) throws ArgumentException {
        String val = getInYml(String.class, instructionMap, keys);
        if (val == null)
            return defaultVal;
        try {
            Method valueOfMethod = defaultVal.getClass().getMethod("valueOf", String.class);
            return (E) valueOfMethod.invoke(null, val);
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException("Should never happen for an enum!", e);
        }
    }


    @Nullable
    private static String findSimpleOptionValue(@Nonnull String[] args,
                                                @Nonnull String shortKey,
                                                @Nonnull String longKey) {
        assert shortKey.length() < longKey.length();
        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.equals("-"+shortKey) || arg.equals("--"+longKey)) {
                if (i < args.length-1) {
                    return args[i+1];
                } else
                    return null;
            }
        }
        return null;
    }

    private static boolean findSimpleOptionPresent(@Nonnull String[] args,
                                                   @Nonnull String shortKey,
                                                   @Nonnull String longKey) {
        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.equals("-"+shortKey) || arg.equals("--"+longKey)) {
                return true;
            }
        }
        return false;
    }

    @Nullable
    private static String findSimpleOptionValueAndRemove(@Nonnull List<String> args,
                                                         @Nonnull String shortKey,
                                                         @Nonnull String longKey) {
        assert shortKey.length() < longKey.length();
        for (int i = 0; i < args.size(); i++) {
            String arg = args.get(i);
            if (arg.equals("-"+shortKey) || arg.equals("--"+longKey)) {
                if (i < args.size()-1) {
                    String res = args.get(i+1);
                    args.remove(i+1);
                    args.remove(i);
                    return res;
                } else
                    return null;
            }
        }
        return null;
    }

    private static boolean findSimpleOptionPresentAndRemove(@Nonnull List<String> args,
                                                            @Nonnull String shortKey,
                                                            @Nonnull String longKey) {
        for (int i = 0; i < args.size(); i++) {
            String arg = args.get(i);
            if (arg.equals("-"+shortKey) || arg.equals("--"+longKey)) {
                args.remove(i);
                return true;
            }
        }
        return false;
    }

    /**
     * Load the file or URL or stream into a string
     *
     * Recognition is done in this way:
     *   - stream: is either stream:stdin, stream:stdout or steam:stderr  (but only stdin makes sense here)
     *   -   "-" is an alias for stream:stdin
     *   - url: starts with http[s]?://     ex: https://example.com/instructions.yml
     *   - file: starts with file:// (typically file://)  or does not match any other source
     *
     * POSSIBLE future extension:
     *   - direct content: starts with direct: and all after that is assumed to be content
     *
     * (if any file clashes with the stream, url or direct scheme, just add file:// in front or use the absolute path)
     *
     * @param contentSource a file, url, or stream.
     * @return the content (== the data read from the content source)
     */
    @Nonnull
    private String loadContent(@Nonnull PrintStream out, @Nonnull PrintStream err, @Nonnull InputStream in,
                               @Nonnull String contentSource) throws ArgumentException {
        InputStream is = null;
        try {
            if (contentSource.equals("stream:stdin") || contentSource.equals("-")) {
                is = in;
            }
            if (contentSource.equals("stream:stderr") || contentSource.equals("stream:stdout")) {
                throw new ArgumentException("Input stream cannot be "+contentSource);
            }

            if (is == null && contentSource.matches("^http[s]?://")) {
                URL url;
                try {
                    url = new URL(contentSource);
                } catch (MalformedURLException e) {
                    throw new ArgumentException("Invalid URL \""+contentSource+"\"", e);
                }
                URLConnection urlConn = url.openConnection();
                urlConn.setReadTimeout(10*1000 /*ms*/);
                is = urlConn.getInputStream();
            }
            if (is == null && contentSource.startsWith("file:")) {
                String filename = contentSource.substring(5);
                try {
                    is = new FileInputStream(filename);
                } catch (FileNotFoundException e) {
                    throw new ArgumentException("Input file \""+filename+"\" not found");
                }
            }
            if (is == null) {
                try{
                    is = new FileInputStream(contentSource);
                } catch (FileNotFoundException e) {
                    throw new ArgumentException("Input file \""+contentSource+"\" not found");
                }
            }

            //noinspection ConstantConditions
            if (is == null) {
                throw new ArgumentException("Unsupported content source \""+contentSource+"\"");
            }

            return IOUtils.streamToString(is, StandardCharsets.UTF_8);
        } catch (IOException e) {
            throw new ArgumentException("IOException while reading \""+contentSource+"\"", e);
        } finally {
            if (is != null)
                try {
                    is.close();
                } catch (IOException e) {
                    //ignore errors closing stream
                }
        }
    }

    /**
     * Basic yml parsing into Map
     *
     * @param ymlContent
     * @return
     */
    public Map parseYml(String ymlContent) throws ArgumentException {
        YAMLFactory yf = new YAMLFactory();
        ObjectMapper mapper = new ObjectMapper(yf);

        Map root;
        try {
            root = mapper.readValue(ymlContent, Map.class);
        } catch (IOException e) {
            throw new ArgumentException("Failed to parse instruction yaml: "+e.getMessage(), e);
        }
        return root;
    }

    public ExperimenterCommandLineParser(String[] args,
                                         final PrintStream out, final PrintStream err, final InputStream in,
                                         BaseCli baseCli) {
        this.args = args;
        this.out = out;
        this.err = err;
        this.in = in;
        this.baseCli = baseCli;
    }

    static public class ProcessInstructionResult {
        public Instruction.Action action;
        public Options options;
        public Instruction instruction;
    }

    /**
     * @param res the ProcessInstructionResult in which to store the result (might be partial result if exception is thrown)
     * @throws CliExitException
     */
    public void processInstruction(ProcessInstructionResult res)
            throws CliExitException {
        LinkedList<String> argList = new LinkedList<>(Arrays.asList(args));
        if (argList.isEmpty()) {
            generalHelp(err);
            throw new RuntimeException("process() should not have been called for no arguments");
        }

        //These arguments are needed early
        String actionCliArg;
        final String ymlInstructionFile;
        final String actionArg = findSimpleOptionValueAndRemove(argList, "a", "action");
        if (actionArg == null) {
            throw new ArgumentException("--action is mandatory");
        }
        Instruction.Action actionArgAction = Instruction.Action.parse(actionArg);
        if (actionArgAction != null) {
            actionCliArg = actionArgAction.name();
            ymlInstructionFile = null;
        } else {
            actionCliArg = null;
            ymlInstructionFile = actionArg;
        }

        //pre-process the yml file, if any
        Map instructionMap;
        if (ymlInstructionFile == null) {
            //no yml file provided: everything is on command line
            instructionMap = new HashMap(); //must be modifyable! We will modify it later.
        } else {
            if (ymlInstructionFile.equals("stream:stdin") || ymlInstructionFile.equals("-")) {
//                if (interactive)
//                    throw new ArgumentException("Cannot use interactive mode when action file is read from stdin.");
            }
            String ymlInstructionContent = loadContent(out, err, in, ymlInstructionFile);
            instructionMap = parseYml(ymlInstructionContent);
        }

        //determine the action  (needed for full yml parsing)
        if (actionCliArg == null) {
            //noinspection ConstantConditions
            if (ymlInstructionFile == null || !instructionMap.containsKey("action")) {
                if (actionCliArg == null) {
                    err.println("ERROR: no action provided. (not in argument list, nor in action file)");
                    generalHelp(err);
                    throw new ArgumentException("no action provided");
                }
                instructionMap.put("action", actionCliArg);
            } else {
                actionCliArg = (String) instructionMap.get("action");
            }
        } else {
            //check if both arguments and yml contain an action. If so, make sure they match.
            //noinspection ConstantConditions
            if (ymlInstructionFile != null && instructionMap.containsKey("action")) {
                String ymlActionString = (String) instructionMap.get("action");
                if (!Instruction.Action.canonize(ymlActionString).equals(Instruction.Action.canonize(actionCliArg))) {
                    throw new ArgumentException("action mismatch between command line argument and action file: " +
                            "\""+ymlActionString+"\" != \""+actionCliArg+"\"");
                }
            }
            instructionMap.put("action", actionCliArg);
        }

        assert instructionMap.containsKey("action");
        assert actionCliArg != null;

        res.action = Instruction.Action.parse(actionCliArg);
        if (res.action == null) {
            err.println("ERROR: unknown action \"" + actionCliArg + "\"");
            generalHelp(err);
            throw new ArgumentException("Unknown action \"" + actionCliArg + "\"");
        }

        res.options = initOptions(res.action);

        //be flexible for action, debug and call lists: allow singleton object instead of list
        for (String outputKey: Arrays.asList("actionOutputs", "debugOutputs", "callOutputs")) {
            if (instructionMap.containsKey(outputKey)) {
                Object out = instructionMap.get(outputKey);
                if (!(out instanceof List)) {
                    instructionMap.put(outputKey, Collections.singletonList(out));
                }
            }
        }

        //this will override the values read from file!
        boolean changed = argumentsToInstruction(res.action, res.options, genericCliArguments, argList.toArray(new String[0]), instructionMap);

        try {
            res.instruction = new InstructionParser().parse(res.action, instructionMap);
        } catch (IllegalArgumentException e) {
            LOG.error("Error in action YML", e);
            throw new CliExitException(Context.ExitReason.BAD_ARGS, "Error in action YML: "+e.getMessage());
        }
        assert res.instruction != null;
        checkMandatory(res.instruction);
    }

    private void checkMandatory(@Nonnull Instruction checkedInstruction) throws ArgumentException {
        assert checkedInstruction != null;
        //TODO: check for missing mandatory fields/combinations
        //Note: the Instruction classes can't check this themselves,
        // as they need to work without all info filled in.
        try {
            if (checkedInstruction instanceof InstructionWithSlice) {
                InstructionWithSlice sliceInstruction = (InstructionWithSlice) checkedInstruction;
                if (!sliceInstruction.getSlice().isValid())
                    throw new ArgumentException("Bad info about slice");
            }
            if (checkedInstruction.getUser() == null) {
                throw new ArgumentException("No user info");
            }
            if (checkedInstruction.getUser().getPem() == null) {
                throw new ArgumentException("No user PEM argument");
            }
            if (checkedInstruction.getUser().getPem().isEmpty()) {
                throw new ArgumentException("Missing user PEM argument");
            }
            switch (checkedInstruction.getAction()) {
                case RUN: {
                    RunInstruction runInstruction = (RunInstruction) checkedInstruction;
                    if (!runInstruction.getExperiment().isValid())
                        throw new ArgumentException("Bad info about experiment");
                    break;
                }
                case CREATESLICE: {
                    CreateSliceInstruction createSliceInstruction = (CreateSliceInstruction) checkedInstruction;
                    break;
                }
                case DELETE: {
                    DeleteInstruction deleteInstruction = (DeleteInstruction) checkedInstruction;
                    break;
                }
                case MANIFEST: {
                    ManifestInstruction manifestInstruction = (ManifestInstruction) checkedInstruction;
                    break;
                }
                case DECRYPTPEM: {
//                    DecryptPemInstruction decryptPemInstruction = (DecryptPemInstruction) instruction;
                    break;
                }
                case SLICECREDENTIAL: {
//                    SliceCredentialInstruction sliceCredentialInstruction = (SliceCredentialInstruction) instruction;
                    break;
                }
                case SLICEINFO: {
                    SliceInfoInstruction sliceInfoInstruction = (SliceInfoInstruction) checkedInstruction;
                    break;
                }
                case USERINFO: {
                    UserInfoInstruction userInfoInstruction = (UserInfoInstruction) checkedInstruction;
                    break;
                }
                case RENEW: {
                    RenewInstruction renewInstruction = (RenewInstruction) checkedInstruction;
                    //TODO: check renew date
                    break;
                }
                case POA: {
                    PoaInstruction poaInstruction = (PoaInstruction) checkedInstruction;
                    //TODO: check PoaAction?
                    break;
                }
            }
        } catch (IllegalArgumentException e) {
            //repackage IllegalArgumentException thrown by isValid() as argument exception
            throw new ArgumentException(e.getMessage(), e);
        }
    }


    private void addGenericOptions(@Nonnull Options options) {
        //instruction file related
        options.addOption(Option.builder("pa").longOpt("print-action")
                .desc("Take no action, but instead output the action file matching the current instructions (which are made up out of cli argument and/or the content of the action file)")
                .build());

        //already processed, and need not be repeated in help
//        options.addOption(Option.builder("a").longOpt("action")
//                .desc("Specify the action name, or the action file (in yml or json format). ")
//                .hasArg()
//                .argName("ACTION NAME | ACTION FILE")
//                .build());

        //user info
        baseCli.addUserConfig(options,
                "Specify the \"context\" file. This old format contains info about the actual login file(s)." +
                        "It is advised to use the -p option instead of this one.",
                true);
//        options.addOption(Option.builder("p").longOpt("pem")
//                .desc("Specify the \"login PEM file\", containing user certificate and private key, in PEM format")
//                .hasArg()
//                .argName("PEM FILE")
//                .build());
//        options.addOption(Option.builder("P").longOpt("password")
//                .desc("cleartext password")
//                .hasArg()
//                .argName("PASSWORD")
//                .build());
//        options.addOption(Option.builder("k").longOpt("key-file")
//                .desc("Specify the file containing the \"login\" SSH private key, in PEM format")
//                .hasArg()
//                .argName("KEY PEM FILE")
//                .build());
        options.addOption(Option.builder("C").longOpt("cert-file")
                .desc("Specify the file containing the \"login\" X509 certificate, in PEM format")
                .hasArg()
                .argName("FILE")
                .build());
//        options.addOption(Option.builder("c").longOpt("context-file")
//                .desc("Specify the \"context\" file. This old format contains info about the actual login file(s)." +
//                        "It is advised to use the -p option instead of this one.")
//                .hasArg()
//                .argName("CONTEXT FILE")
//                .build());

//        options.addOption("i", "interactive", false, "possibly request some missing options interactively");

        //output, version and help
        options.addOption("q", "quiet", false, "less output");
        options.addOption(null, "silent", false, "less output (same as --quiet)");
        options.addOption("d", "debug", false, "extra debugging output");
        options.addOption("v", "version", false, "show version and exit");
        options.addOption("h", "help", false, "show help and exit");
    }

    @Nonnull
    private Options initOptions(@Nonnull Instruction.Action action) {
        Options res = new Options();

        addGenericOptions(res);

//        bindArg(res, "S", "speaks-for",
//                "SpeaksFor Credential file",
//                "SPEAKS FOR CREDENTIAL FILE",
//                "user", "speaksFor");
//
//
        if (action.isSliceArgSupported())
            res.addOption(Option.builder("s").longOpt("slice")
                    .desc("The URN or name of the slice to use. (auto detected)")
                    .hasArg()
                    .argName("SLICE URN/NAME")
                    .build());//todo: how to handle that this is potentially both slice.name and slice.urn ?

//       TODO: add equivalent res:   BaseCli.addAuthoritiesConfig(res);

//        res.addOption(Option.builder().longOpt("speaks-for")
//                .desc("The speaksFor certificate (in a file), enabling you make calls on behalf of another user.")
//                .hasArg()
//                .argName("CERTIFICATE FILENAME")
//                .build());
        if (action.isSliceArgSupported())
            res.addOption(Option.builder("S").longOpt("project-name")
                    .desc("The name of the project (= sub authority) of the slice. This is an optional argument (however some authorities might require a project!). " +
                            "You can also use \"" + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + "\" as project name, in that case, the last (determined by your SA) project you are a member of will be used.")
                    .hasArg()
                    .argName("NAME")
                    .build());
        if (action == Instruction.Action.RUN || action == Instruction.Action.RENEW || action == Instruction.Action.CREATESLICE) {
            res.addOption(Option.builder().longOpt("expiration-date")
                    .desc("The date and time at which this slice expires. In RFC3339 format. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .argName("RFC3339 DATE")
                    .build());
            res.addOption(Option.builder("e").longOpt("expiration-hours")
                    .desc("The number of hours after which this slice expires. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .argName("INTEGER")
                    .build());
        }
        if (action == Instruction.Action.RUN || action == Instruction.Action.CREATESLICE) {
            res.addOption(Option.builder().longOpt("share-slice")
                    .desc("List of users to share slice with. Either use the URN of each user, or the short username. " +
                            "You can also specify the special value \"PROJECTUSERS\", which will automatically share the slice with all users in the project. " +
                            "Multiple users can be specified by separating them with a comma.")
                    .hasArg()
                    .argName("USERNAME(S)")
                    .build());
        }
        if (action == Instruction.Action.SLICEINFO ||
                action == Instruction.Action.USERINFO ||
                action == Instruction.Action.CREATESLICE ||
                action == Instruction.Action.MANIFEST ||
                action == Instruction.Action.SLICECREDENTIAL)
            res.addOption(Option.builder().longOpt("output-format")
                    .desc("The output format to use (for user information). Choices: \"text\" or \"json\". Default:json")
                    .hasArg()
                    .argName("FORMAT")
                    .build());
        if (action == Instruction.Action.POA) {
            String actionOptionsString = actionOptions.stream().collect(Collectors.joining(", "));
            res.addOption(Option.builder("poa").longOpt("poa-action")
                    .desc("The operational action to perform. (Options: " + actionOptionsString + ")")
                    .hasArg()
                    .argName("ACTION")
                    .build());
            res.addOption(Option.builder("t").longOpt("target-sliver")
                    .desc("The URN of the sliver to perform the action on.")
                    .hasArg()
                    .argName("SLIVER URN")
                    .build());
        }
        if (action == Instruction.Action.RUN)
            res.addOption(Option.builder("r").longOpt("rspec")
                    .desc("The rspec file to use for creating a sliver")
                    .hasArg()
                    .argName("RSPEC XML FILE")
                    .build());

        //deprecated option:
//        if (action == Instruction.Action.SLICEINFO || action == Instruction.Action.RENEW || action == Instruction.Action.POA)
//            res.addOption(Option.builder("r").longOpt("rspec")
//                    .desc("The rspec file to get the needed authority information from (manifest or request)  (not mandatory, but might not work without)")
//                    .hasArg()
//                    .argName("RSPEC XML FILE")
//                    .build());

        //ansible options:
//        if (action == Instruction.Action.CREATE || action == Instruction.Action.MANIFEST)
//            res.addOption(Option.builder().longOpt("ansible-dir")
//                    .desc("The dir in which to save ansible config files. Will be created if it doesn't exist. Will overwrite any existing files.")
//                    .hasArg()
//                    .argName("DIR")
//                    .build());
//        if (action == Instruction.Action.CREATE || action == Instruction.Action.MANIFEST) {
//            res.addOption(Option.builder().longOpt("ansible-execute-rspec-playbooks")
//                    .desc("Execute all ansible playbooks found within request Rspec " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() +
//                            " elements. (default: ignore them)")
//                    .build());
//            res.addOption(Option.builder().longOpt("ansible-add-playbook")
//                    .desc("Add an ansible playbook to execute. The argument specifies the source, " +
//                            "and optionally (after a space) the target (this option can be specified multiple times)")
//                    .numberOfArgs(2)
//                    .argName(RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + " [" +
//                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + "]")
//                    .build());
//            res.addOption(Option.builder().longOpt("ansible-allow-playbook-inputfile")
//                    .desc("Allow local files in " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
//                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + ". (default: only allow URLs as " +
//                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + ")")
//                    .build());
//            res.addOption(Option.builder().longOpt("ansible-allow-any-playbook-outputfile")
//                    .desc("Allow any local file as " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
//                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + ". (default: only use the basename of the specified " +
//                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + ", and use the --ansible-dir as directory)")
//                    .build());
//            res.addOption(Option.builder().longOpt("ansible-playbook-exe")
//                    .desc("Executable for the ansible playbook. Default is \"ansible-playbook\" and is usually fine. " +
//                            "On some systems might need to be \"/usr/bin/python2 /usr/bin/ansible-playbook\"")
//                    .hasArg()
//                    .build());
//            res.addOption(Option.builder().longOpt("ansible-debug")
//                    .desc("Set debug flags when calling ansible playbook.")
//                    .build());
//        }
        res.addOption(Option.builder().longOpt("manifest")
                .desc("The file in which the manifest must be stored. (default: manifest-<SLICENAME>.rspec)")
                .hasArg()
                .argName("FILE")
                .build());
        if (action == Instruction.Action.DELETE)
            res.addOption(Option.builder("r").longOpt("rspec")
                    .desc("The rspec file to get the needed authority information from (manifest or request) (not mandatory, but might not work without)")
                    .hasArg()
                    .argName("RSPEC XML FILE")
                    .build());
        if (action == Instruction.Action.RUN) {
            res.addOption(Option.builder().longOpt("delete-on-create-failure")
                    .desc("If there is a failure in a call made to create the sliver(s), delete all resources everywhere before exiting.")
                    .build());
            res.addOption(Option.builder().longOpt("delete-on-become-ready-failure")
                    .desc("If there is a failure in a call while waiting for the sliver(s) to become ready, delete all resources everywhere before exiting.")
                    .build());
            res.addOption(Option.builder().longOpt("abort-if-slivers-exist")
                    .desc("Do not create any slivers if any sliver already exists on any of the authorities.")
                    .build());
            res.addOption(Option.builder().longOpt("create-slice")
                    .desc("Create the slice if it doesn't exist. (if it exists, this option is ignored)")
                    .build());
            res.addOption(Option.builder().longOpt("rewrite-rspec")
                    .desc("Parse the provided RSpec, and reconstruct it again, before sending it to the server. This can help prevent some server sides bug caused by valid but atypical RSpecs.")
                    .build());
            res.addOption(Option.builder().longOpt("bind-rspec")
                    .desc("Bind any unbound nodes in the RSpec to a specified authority. This leaves bound nodes as they are. (this imples --rewrite-rspec)")
                    .hasArg()
                    .argName("COMPONENT MANAGER URN")
                    .build());
            res.addOption(Option.builder("k").longOpt("ssh-keys")
                    .desc("Specify which ssh keys to add. The argument is a (comma separated) list of res.\n" +
                            "Available res: usercert,userkeys,rspec,shareduserallkeys\n" +
                            "   usercert: add the user making the calls, with the ssh key from the certificate in the login PEM file.\n" +
                            "   userkeys: add the user making the calls, with the ssh keys that are stored on the user authority MA server. (usercert and userkeys can be combined to add all keys)\n" +
                            "   shareduserallkeys: add the ssh keys that are stored on the user authority MA server for the users the slice is shared with AND the ssh keys from that user's login certificate.\n" +
                            "   rspec: add the users and keys specified in the RSpec itself.\n" +
                            "This option is optional, default: usercert,rspec,shareduserallkeys")
                    .hasArg()
                    .argName("OPTION LIST")
                    .build());
//            res.addOption(Option.builder().longOpt("slice-recover-info")
//                    .desc("A file into which slice recover info will be stored. This can be used by the jFed GUI (if copied to the correct directory).")
//                    .hasArg()
//                    .argName("FILE")
//                    .build());
            res.addOption(Option.builder().longOpt("nowait")
                    .desc("Do not wait until sliver is ready (default: wait until ready)")
                    .build());
        }
        res.addOption(Option.builder().longOpt("call-log")
                .desc("A file into which all call details will be stored")
                .hasArg()
                .argName("FILE")
                .build());
        res.addOption(Option.builder().longOpt("print-calls")
                .desc("Print all calls to stdout")
                .build());
//        res.addOption(Option.builder("am").longOpt("am-api")
//                .desc("The AM Api version to use (\"2\" and \"3\" supported, default: choose automatically)")
//                .hasArg()
//                .argName("AM API VERSION")
//                .build());
//        res.addOption(Option.builder().longOpt("fake")
//                .desc("Do everything, expect actually making the slice and sliver calls. Useful for debugging syntax. (will make calls relating to retrieving user data)")
//                .build());
//        if (action != Instruction.Action.CREATE_SLICE && action != Instruction.Action.SHOW_SLICE_INFO)
//            res.addOption(Option.builder().longOpt("fake-sliver")
//                    .desc("Do everything, expect actually making the sliver calls. (will make calls relating to retrieving user data and creating or retrieving slice)")
//                    .build());


        return res;
    }

    private static void deepMapHelper(Map target, Object newVal, String... keys) {
        assert keys.length > 0;
        for (int i = 0; i < keys.length-1; i++) {
            String key = keys[i];
            target = (Map) target.computeIfAbsent(key, (k) -> new HashMap());
        }
        target.put(keys[keys.length-1], newVal);
    }
    private static List<Object> deepMapListHelper(Map target, String... keys) {
        assert keys.length > 0;
        for (int i = 0; i < keys.length-1; i++) {
            String key = keys[i];
            target = (Map<String, Object>) target.computeIfAbsent(key, (k) -> new HashMap<String, Object>());
        }
        return (List) target.computeIfAbsent(keys[keys.length - 1], (k) -> new ArrayList<>());
    }

    /**
     *
     * @param action
     * @param options
     * @param genericCliArguments
     * @param args
     * @param instructionMap
     * @return true if the instructionMap was changed
     * @throws ArgumentException
     */
    private boolean argumentsToInstruction(@Nonnull Instruction.Action action,
                                           @Nonnull Options options,
                                           @Nonnull GenericCliArguments genericCliArguments,
                                           @Nonnull String[] args,
                                           @Nonnull Map instructionMap) throws ArgumentException {
        boolean changed = false;
        CommandLineParser parser = new BasicParser();
        CommandLine line;
        try {
            line = parser.parse(options, args);
        } catch (ParseException e) {
            throw new ArgumentException("Argument error: "+e.getMessage(), e);
        }

//                findSimpleOptionValue(args, "p", "cert-and-key-file"),
//                findSimpleOptionValue(args, "k", "key-file"),
//                findSimpleOptionValue(args, "C", "cert-file"),
//                findSimpleOptionValue(args, "c", "context-file"),
//                findSimpleOptionValue(args, "P", "private-key-password"),

        String pemArg = line.getOptionValue("cert-and-key-file");
        String keyArg = line.getOptionValue("key-file");
        String certArg = line.getOptionValue("cert-file");
        String contextArg = line.getOptionValue("context-file");
        String passArg = line.getOptionValue("private-key-password");

        if (pemArg != null) {
            deepMapListHelper(instructionMap, "user", "pem")
                    .add(pemArg);
            changed = true;
        }
        if (keyArg != null) {
            deepMapListHelper(instructionMap, "user", "pem")
                    .add(keyArg);
            changed = true;
        }
        if (certArg != null) {
            deepMapListHelper(instructionMap, "user", "pem")
                    .add(certArg);
            changed = true;
        }
        if (passArg != null) {
            deepMapHelper(instructionMap, "DIRECT", "user", "passwordMethod");
            deepMapHelper(instructionMap, passArg, "user", "password");
            changed = true;
        }
        if (contextArg != null) {
            //option ignored for this processing. It is hacked in amnually using the BaseCli initUserModule
            err.println("--context-file is deprecated, but will still work for now.");
        }



        String sliceIdArg = line.getOptionValue("slice");
        String projectNameArg = line.getOptionValue("project-name");
        String sliceExpirationDateArg = line.getOptionValue("expiration-date");
        String sliceExpirationHoursArg = line.getOptionValue("expiration-hours");
        String rspecArg = line.getOptionValue("rspec");
        String bindRspecArg = line.getOptionValue("bind-rspec");
//        boolean rewriteRspecArg = line.hasOption("rewrite-rspec") || bindRspecArg != null;
        String shareSliceUsersArg = line.getOptionValue("share-slice");
        @Nonnull String outputFormatArg = line.getOptionValue("output-format") == null ? null : line.getOptionValue("output-format");
        String amApiVersionArg = line.getOptionValue("am-api");
        String sshKeysArg = line.getOptionValue("ssh-keys");
        String manifestArg = line.getOptionValue("manifest");
        String callLogArg = line.getOptionValue("call-log");
        String sliceRecoverInfoArg = line.getOptionValue("slice-recover-info");
        boolean createSliceArg = line.hasOption("create-slice");
//        boolean stitchingArg = line.hasOption("stitching");
//        boolean debug = line.hasOption("debug");
        boolean printCalls = line.hasOption("print-calls");
        boolean fake = line.hasOption("fake");
//        boolean logbackLogging = line.hasOption("logging");
//        boolean fakeSliver = line.hasOption("fake-sliver");
        boolean waitUntilReady = !line.hasOption("nowait");
//        boolean silent = !debug && (line.hasOption("quiet") || line.hasOption("silent"));
        boolean abortIfSliversExist = line.hasOption("abort-if-slivers-exist");
        boolean deleteOnCreateFailure = line.hasOption("--delete-on-create-failure");
        boolean deleteOnBecomeReadyFailure = line.hasOption("--delete-on-become-ready-failure");
        String[] speaksForArgs = line.getOptionValues("speaks-for");
        String poaActionArg = line.getOptionValue("poa-action");
        String poaTargetSliverUrnArg = line.getOptionValue("target-sliver");

        String ansibleDirName = line.getOptionValue("ansible-dir");
        String ansiblePlayBookExe = line.getOptionValue("ansible-playbook-exe", "ansible-playbook");
        boolean ansibleExecuteRspecPlaybooks = line.hasOption("ansible-execute-rspec-playbooks");
        String[] ansibleExtraPlaybooks = line.getOptionValues("ansible-add-playbook");
        boolean noPlaybookToExecute = !ansibleExecuteRspecPlaybooks && (ansibleExtraPlaybooks == null || ansibleExtraPlaybooks.length == 0);
        boolean ansibleAllowPlaybookInputfile = line.hasOption("ansible-allow-playbook-inputfile");
        boolean ansibleAllowAnyPlaybookOutputFile = line.hasOption("ansible-allow-any-playbook-outputfile");
        boolean ansibleDebug = line.hasOption("ansible-debug");

        if (sliceIdArg != null) {
            if (sliceIdArg.startsWith("urn:publicid:IDN")) {
                sliceIdArg = GeniUrn.parse(sliceIdArg).getResourceName();
            }
            if (action == Instruction.Action.RUN) {
                deepMapHelper(instructionMap, sliceIdArg, "experiment", "slice", "sliceName");
            } else {
                deepMapHelper(instructionMap, sliceIdArg, "slice", "sliceName");
            }
            changed = true;
        }
        if (projectNameArg != null) {
            if (projectNameArg.equalsIgnoreCase(CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
                if (action == Instruction.Action.RUN) {
                    deepMapHelper(instructionMap, Slice.ProjectSource.AUTO.name(), "experiment", "slice", "projectSource");
                } else {
                    deepMapHelper(instructionMap, Slice.ProjectSource.AUTO.name(), "slice", "projectSource");
                }
            } else {
                if (action == Instruction.Action.RUN) {
                    deepMapHelper(instructionMap, Slice.ProjectSource.PROVIDED.name(), "experiment", "slice", "projectSource");
                    deepMapHelper(instructionMap, projectNameArg, "experiment", "slice", "project");
                } else {
                    deepMapHelper(instructionMap, Slice.ProjectSource.PROVIDED.name(), "slice", "projectSource");
                    deepMapHelper(instructionMap, projectNameArg, "slice", "project");
                }
            }
            changed = true;
        }
        if (sliceExpirationDateArg != null) {
            Instant requestedExpirationDate;
            try {
                requestedExpirationDate = RFC3339Util.rfc3339StringToInstant(sliceExpirationDateArg);
            } catch (java.text.ParseException e) {
                throw new ArgumentException("Specified slice expiration date (\"" + sliceExpirationDateArg + "\") is not in RFC3339 format. The current date in RFC3339 format is: " +
                        RFC3339Util.dateToRFC3339String(new Date()));
            }
            Instant now = Instant.now();
            long expireTimeMin = Duration.between(requestedExpirationDate, now).toMinutes();

            if (action == Instruction.Action.RUN) {
                deepMapHelper(instructionMap, expireTimeMin, "experiment", "slice", "expireTimeMin");
            } else {
                deepMapHelper(instructionMap, expireTimeMin, "slice", "expireTimeMin");
            }
            changed = true;
        }
        if (sliceExpirationHoursArg != null) {
            long expireTimeMin;
            try {
                int hours = Integer.parseInt(sliceExpirationHoursArg);
                expireTimeMin = 60 * hours;
            } catch (NumberFormatException e) {
                throw new ArgumentException("Specified slice expiration hours (\"" + sliceExpirationHoursArg + "\") is not a valid number.");
            }
            if (action == Instruction.Action.RUN) {
                deepMapHelper(instructionMap, expireTimeMin, "experiment", "slice", "expireTimeMin");
            } else {
                deepMapHelper(instructionMap, expireTimeMin, "slice", "expireTimeMin");
            }
            changed = true;
        }
        if (rspecArg != null) {
            if (rspecArg.startsWith("http")) {
                deepMapHelper(instructionMap, "PROVIDE_URL", "experiment", "requestRSpec", "source");
            } else {
                if (rspecArg.startsWith("file:"))
                    rspecArg = rspecArg.substring(5);
                deepMapHelper(instructionMap, "PROVIDE_FILE", "experiment", "requestRSpec", "source");
            }
            deepMapHelper(instructionMap, rspecArg, "experiment", "requestRSpec", "providedContentSource");
        }
        if (bindRspecArg != null) {
            deepMapHelper(instructionMap, true, "experiment", "requestRSpec", "bindTestedServerToUnboundNodes");
            deepMapHelper(instructionMap, bindRspecArg, "experiment", "requestRSpec", "bindUnboundNodesUrn");
        }
        if (line.hasOption("rewrite-rspec")) {
            err.println("The --rewrite-rspec option is deprecated and will be ignored");
        }
        if (line.hasOption("am-api")) {
            err.println("The --am-api option is deprecated and will be ignored");
        }
        if (shareSliceUsersArg != null) {
            List<String> shareWithUsers = new ArrayList<>();
            boolean shareSliceWithProjectMembers = false;
            String[] users = shareSliceUsersArg.split(",");
            for (String opt : users) {
                boolean validOpt = false;

                if (Objects.equals(opt, "PROJECTUSERS")) {
                    shareSliceWithProjectMembers = true;
                    validOpt = true;
                }
                if (!validOpt && GeniUrn.valid(opt)) {
                    GeniUrn userUrn = GeniUrn.parse(opt);
                    assert userUrn != null;
                    shareWithUsers.add(opt);
                    validOpt = true;
                }
                if (!validOpt && opt.matches("[a-zA-Z][a-zA-Z0-9]*") && opt.length() <= 8) {
//                    String tla = targetUserUrn.getEncodedTopLevelAuthority();
//                    GeniUrn userUrn = GeniUrn.createGeniUrnFromEncodedParts(tla, "user", opt);
//                    shareUserUrns.add(userUrn);
                    shareWithUsers.add(opt);
                    validOpt = true;
                }

                if (!validOpt) {
                    throw new ArgumentException("The share-slice option you have specified (" + opt + ") is not valid a valid URN or username");
                }
            }

            deepMapHelper(instructionMap, shareSliceWithProjectMembers, "shareWith", "projectMembers");
            deepMapHelper(instructionMap, shareWithUsers, "shareWith", "users");
        }
        if (outputFormatArg != null) {
            ActionOutput actionOutput = new ActionOutput(
                    ActionOutput.Format.parse(outputFormatArg),
                    OutputTarget.STDOUT, null);
            deepMapListHelper(instructionMap, "actionOutputs")
                    .add(actionOutput);
        }
        if (sshKeysArg != null) {
            String[] opts = sshKeysArg.split(",");

            boolean sshKeysUserCert = false;
            boolean sshKeysUserServerKeys = false;
            boolean sshKeysSharedUserKeys = false;
            boolean sshKeysRspec = false;

            for (String opt : opts) {
                boolean validOpt = false;

                if (opt.equalsIgnoreCase("usercert")) {
                    sshKeysUserCert = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("userkeys")) {
                    sshKeysUserServerKeys = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("shareduserallkeys")) {
                    sshKeysSharedUserKeys = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("rspec")) {
                    sshKeysRspec = true;
                    validOpt = true;
                }

                if (!validOpt) {
                    throw new ArgumentException("The ssh-keys option you have specified (" + opt + ") is not valid");
                }
            }

            deepMapHelper(instructionMap, sshKeysUserCert, "shareWith", "userCert");
            deepMapHelper(instructionMap, sshKeysUserServerKeys, "shareWith", "userKeys");
            deepMapHelper(instructionMap, sshKeysSharedUserKeys, "shareWith", "shareWith");
            deepMapHelper(instructionMap, sshKeysRspec, "shareWith", "rspec");
//            deepMapHelper(instructionMap, sshKeysUserCert, "shareWith", "extraUsernames");
//            deepMapHelper(instructionMap, sshKeysUserCert, "shareWith", "extraKeys");
        }
        if (manifestArg != null) {
            ActionOutput actionOutput = new ActionOutput(
                    ActionOutput.Format.TXT,
                    OutputTarget.FILE, manifestArg);
            deepMapListHelper(instructionMap, "actionOutputs")
                    .add(actionOutput);
        }
        if (callLogArg != null && outputFormatArg != null) {
            deepMapListHelper(instructionMap, "callOutputs").add(
                    new CallOutput(CallOutput.Format.valueOf(outputFormatArg.toUpperCase()), OutputTarget.FILE, callLogArg)
            );
        }
        if (line.hasOption("slice-recover-info")) {
            err.println("The --slice-recover-info option is deprecated and will be ignored");
        }
        if (line.hasOption("fake")) {
            throw new ArgumentException("The --fake option is deprecated. Will not continue.");
        }
        if (createSliceArg) {
            if (action == Instruction.Action.RUN) {
                deepMapHelper(instructionMap, false, "experiment", "slice", "failOnNonExistingSlice");
                deepMapHelper(instructionMap, true, "experiment", "slice", "failOnExistingSlice");
            } else {
                deepMapHelper(instructionMap, false, "slice", "failOnNonExistingSlice");
                deepMapHelper(instructionMap, true, "slice", "failOnExistingSlice");
            }
        }
        if (printCalls) {
            deepMapListHelper(instructionMap, "callOutputs").add(
                    new CallOutput(CallOutput.Format.TEXT, OutputTarget.STDOUT, callLogArg)
            );
        }
        if (line.hasOption("logging")) {
            err.println("The --logging option is deprecated and will be ignored");
        }
        if (line.hasOption("fake-sliver")) {
            err.println("The --fake-sliver option is deprecated and will be ignored");
        }
        if (line.hasOption("nowait")) {
            deepMapHelper(instructionMap, false, "experiment", "waitForReady", "enabled");
        }
        if (abortIfSliversExist) {
//            deepMapHelper(instructionMap, false, "experiment", "waitForReady", "enabled");
            err.println("The --abort-if-slivers-exist option is deprecated and will be ignored");
        }
        if (deleteOnCreateFailure) {
            deepMapHelper(instructionMap, true,  "deleteOn", "failCreate");
        }
        if (deleteOnBecomeReadyFailure) {
            deepMapHelper(instructionMap, true,  "deleteOn", "failBecomeReady");
        }
        if (line.hasOption("speaks-for")) {
            throw new ArgumentException("The --speaks-for option is deprecated. Will not continue.");
        }

        if (poaActionArg != null) {
            deepMapHelper(instructionMap, PoaInstruction.PoaAction.parse(poaActionArg),  "poaAction");
        }
        if (poaTargetSliverUrnArg != null) {
            deepMapHelper(instructionMap, poaTargetSliverUrnArg,  "sliverUrn");
        }

        return changed;
    }




    public Instruction.Action getCommand() {
        return instruction.getAction();
    }
    /** returns true if the user provided Instruction has an Action matching is any of the given Actions */
    public boolean isAction(Instruction.Action ... anyActions) {
        for (Instruction.Action a : anyActions) {
            if (Objects.equals(instruction.getAction(), a))
                return true;
        }
        return false;
    }

    public void showVersion() {
        showVersion(out);
    }
    public static void showVersion(PrintStream out) {
        out.println("jFed Experimenter CLI 2 -- http://jfed.iminds.be/");
        out.println("Send bugreports to: jfed-bugreports@atlantis.ugent.be");

        JFedVersionInfo jFedVersionInfo = new JFedVersionInfo();
        out.println("Version: " + jFedVersionInfo.getFullVersionString());
        out.println("Environment: "+jFedVersionInfo.getEnvironmentString());
    }

    @Nonnull
    public static String getAllActions() {
        return Arrays.stream(Instruction.Action.values())
                .map(Instruction.Action::getNiceName)
                .collect(Collectors.joining(" "));
    }

    public static void generalHelp(final PrintStream err) {
        err.println("Syntax: jfed-experimenter-cli [-a|--action <ACTION NAME | ACTION FILE>] [action_options ... ]\n" +
                "Available Actions: "+ getAllActions()+"\n\n"+
                "More help:\n" +
                "   Specific help: jfed-experimenter-cli --action <ACTION NAME> --help"+"\n"+
                "   Documentation: https://doc.ilabt.imec.be/jfed-documentation-5.9/otherjfedtools.html#experimenter-cli-2\n");
    }


    @Nonnull
    private GenericCliArguments processGenericCliArguments() {
        return new GenericCliArguments(
                findSimpleOptionValue(args, "c", "context-file"),
                findSimpleOptionPresent(args, "s", "silent"),
                findSimpleOptionPresent(args, "d", "debug"),
                findSimpleOptionPresent(args, "h", "help"),
                findSimpleOptionPresent(args, "w", "wizard"),
                findSimpleOptionPresent(args, "v", "version"),
                findSimpleOptionPresent(args, "pa", "print-action"),
                findSimpleOptionPresent(args, "a", "action")
        );
    }

    @Nonnull
    public GenericCliArguments getGenericCliArguments() {
        if (genericCliArguments == null) {
            genericCliArguments = processGenericCliArguments();
        }
        return genericCliArguments;
    }

    public static class GenericCliArguments {
        @Nullable private final String contextFilename;
        private final boolean silent;
        private final boolean debug;
        private final boolean help;
        private final boolean version;
        private final boolean wizard;
        private final boolean printAction;
        private final boolean hasActionArg;

        public GenericCliArguments(
                @Nullable String contextFilename,
                boolean silent, boolean debug,
                boolean help, boolean wizard, boolean version, boolean printAction,
                boolean hasActionArg) {
//            this.pemKeyCertFilename = pemKeyCertFilename;
//            this.keyFilename = keyFilename;
//            this.certFilename = certFilename;
            this.contextFilename = contextFilename;
//            this.privateKeyPassword = privateKeyPassword;
//            this.interactive = interactive;
            this.silent = silent;
            this.debug = debug;
            this.help = help;
            this.wizard = wizard;
            this.version = version;
            this.printAction = printAction;
            this.hasActionArg = hasActionArg;
        }

        @Nullable
        public String getContextFilename() {
            return contextFilename;
        }

        public boolean isSilent() {
            return silent;
        }

        public boolean isDebug() {
            return debug;
        }

        public boolean isHelp() {
            return help;
        }

        public boolean isVersion() {
            return version;
        }

        public boolean isWizard() {
            return wizard;
        }

        public boolean isPrintAction() {
            return printAction;
        }

        public boolean hasAction() {
            return hasActionArg;
        }
    }
}
